/*****************************************************************************/
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 11/13/2001                                                 */
/*                                                                           */
/* File: "control.c"                                                         */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "manip.h"
#include "control.h"
#include "robots.h"

extern Database Table[];

/*****************************************************************************/

int pid_control(double *tau, CONTROLLER *c)
{
  int i;
  
  for(i = 0; i < c->n; i++) {
    tau[i] = c->kp[i]*c->err[i] + c->kv[i]*c->derr[i] + c->ki[i]*c->ierr[i];
    c->ierr[i] += c->h*c->err[i];			 
  }
	
  return(1);
}

/*****************************************************************************/

int pid_control_with_ddthd(double *tau, CONTROLLER *c)
{
  int i;
  
  for(i = 0; i < c->n; i++) {
    tau[i] = c->ddthd[i] + c->kp[i]*c->err[i] + c->kv[i]*c->derr[i] + 
      c->ki[i]*c->ierr[i];
    c->ierr[i] += c->h*c->err[i];			 
  }
	
  return(1);
}

/*****************************************************************************/

int pid_control_with_gravity_compensation(double *tau, CONTROLLER *c)
{
  int i;
  double g[NMAX];

  c->model(NULL, NULL, &g[0], &c->th[0], &c->dth[0]);

  for(i = 0; i < c->n; i++) {
    tau[i] = c->ddthd[i] + c->kp[i]*c->err[i] + c->kv[i]*c->derr[i] + 
      c->ki[i]*c->ierr[i] + g[i];
    c->ierr[i] += c->h*c->err[i];			 
  }
	
  return(1);
}

/*****************************************************************************/

int partitioned_control(double *tau, CONTROLLER *c)
{
  int i, j;
  double taup[NMAX], g[NMAX], v[NMAX], m[NMAX*NMAX];

  c->model(&m[0], &v[0], &g[0], &c->th[0], &c->dth[0]);

  for(i = 0; i < c->n; i++) {
    taup[i] = c->ddthd[i] + c->kp[i]*c->err[i] + c->kv[i]*c->derr[i] + 
      c->ki[i]*c->ierr[i];
    c->ierr[i] += c->h*c->err[i];			 
  }

  for(i = 0; i < c->n; i++) {
    tau[i] = 0.;
    for(j = 0; j < c->n; j++) {
      tau[i] += m[c->n*i + j]*taup[j];
    }
    tau[i] += v[i] + g[i];
  }
	
  return(1);
}

/*****************************************************************************/

int set_up_controller(MANIP_PARAMETERS *p)
{
  int index;

  index = p->robot_type - 1;

  p->n_links = Table[index].n_links;
  p->robot = Table[index].robot;
  p->control = pid_control;

  switch(p->controller_type)
    {
    case 1: /* simple PID control */
      {
	p->control = pid_control;
	break;
      }
    case 2: /* PID control with desired acceleration term */
      {
	p->control = pid_control_with_ddthd;
	break;
      }
    case 3: /* PID control with gravity compensation and perfect model */
      {
	p->model = Table[index].actual_model;
	p->control = pid_control_with_gravity_compensation;
	break;
      }
    case 4: /* PID control with gravity compensation and idealized model */
      {
	p->model = Table[index].assumed_model;
	p->control = pid_control_with_gravity_compensation;
	break;
      }
    case 5: /* Partitioned control with perfect model */
      {
	p->model = Table[index].actual_model;
	p->control = partitioned_control;
	break;
      }
    case 6: /* Partitioned control with imperfect model */
      {
	p->model = Table[index].assumed_model;
	p->control = partitioned_control;
	break;
      }
    }
  
  return(1);
}

/*****************************************************************************/


