/*****************************************************************************/
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 11/13/2001                                                 */
/*                                                                           */
/* File: "manip.c"                                                           */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "manip.h"
#include "control.h"
#include "robots.h"

/*****************************************************************************/

int file_exist(char *file_name)
{
  int flag;
  FILE *fp;

  fp = fopen(file_name, "r");

  if(fp != NULL) {
    flag = 1;
    fclose(fp);
  } else
    flag = 0;

  return(flag);
}

/*****************************************************************************/

int file_no_exist(char *file_name, char *type)
{
  if(file_exist(file_name) == 0) {
    if(file_name == NULL)
      fprintf(stderr, "-- No %s file given. Exiting.\n", type);
    else 
      fprintf(stderr, "-- The %s file (%s) doesn't exist. Exiting.\n", 
              type, file_name);
    exit(-1);
  }

  return(1);
}

/*****************************************************************************/

void display_manip_parameters(FILE *fp, MANIP_PARAMETERS *p)
{
  int i;

  fprintf(fp, "----------------------------------------\n");
  fprintf(fp, "Number of links = %d.\n", p->n_links);
  fprintf(fp, "Robot type = %d.\n", p->robot_type);
  fprintf(fp, "Controller type = %d.\n", p->controller_type);
  fprintf(fp, "Controller freq = %.1f.\n", p->frequency);
  fprintf(fp, "Total sim. time = %.1f.\n", p->total_time);
  fprintf(fp, "Start configuration: time = %.1f\n    ", p->start.time);
  for(i = 0; i < p->n_links; i++)
    fprintf(fp, " %.6f", p->start.joints[i]);
  fprintf(fp,  "\n");

  fprintf(fp, "Goal configuration: time = %.1f\n     ", p->goal.time);
  for(i = 0; i < p->n_links; i++)
    fprintf(fp, " %.6f", p->goal.joints[i]);
  fprintf(fp,  "\n");

  fprintf(fp, "KP gains:\n    ");
  for(i = 0; i < p->n_links; i++)
    fprintf(fp, " %.3f", p->KP[i]);
  fprintf(fp,  "\n");

  fprintf(fp, "KV gains:\n    ");
  for(i = 0; i < p->n_links; i++)
    fprintf(fp, " %.3f", p->KV[i]);
  fprintf(fp,  "\n");

  fprintf(fp, "KI gains:\n    ");
  for(i = 0; i < p->n_links; i++)
    fprintf(fp, " %.3f", p->KI[i]);
  fprintf(fp,  "\n");
  fprintf(fp, "----------------------------------------\n");

  return;
}

/*****************************************************************************/

void initialize_manip_parameters(MANIP_PARAMETERS *p)
{
  int i;

  p->n_links = 2;
  p->controller_type = 1;
  p->robot_type = 1;
  p->frequency = 100.;
  p->step_size = 1./p->frequency;
  p->total_time = 10.;
  p->start.time = 0.;
  p->goal.time = 5.;

  for(i = 0; i < NMAX; i++) {
    p->KP[i] = 0.;
    p->KV[i] = 0.;
    p->KI[i] = 0.;
    p->start.joints[i] = 0.;
    p->goal.joints[i] = 0.;
  }

  p->robot = robot_actual_2link_pointmass;

  return;
}

/*****************************************************************************/

void get_manip_parameters(char *file_name, MANIP_PARAMETERS *p)
{
  FILE *fp;
  char line[256];

  file_no_exist(file_name, "parameter");

  fp = fopen(file_name, "r");

  while(fscanf(fp, "%[^\n]\n", &line[0]) != EOF) {
    process_manip_line(line, p, 1);
  }
  
  fclose(fp);

  return;
}

/*****************************************************************************/

int process_manip_line(char *line, MANIP_PARAMETERS *p, int alert_flag)
{
  char *separators = " \t\v\f\r\n,", *first;

  if(line[0] == '#' || line[0] == '\n')
    return(1);

  first = strtok(line, separators);

  if(strcmp(first, "KP") == 0) {

    read_all_line_values(&p->KP[0], separators);
    
  } else if (strcmp(first, "KV") == 0) {

    read_all_line_values(&p->KV[0], separators);

  } else if (strcmp(first, "KI") == 0) {

    read_all_line_values(&p->KI[0], separators);
    
  } else if (strcmp(first, "Controller") == 0) {

    p->controller_type = atoi(strtok(NULL, separators));

  } else if (strcmp(first, "Robot") == 0) {

    p->robot_type = atoi(strtok(NULL, separators));

  } else if (strcmp(first, "Start") == 0) {

    read_via_point(&p->start, separators);

  } else if (strcmp(first, "Goal") == 0) {

    p->n_links = read_via_point(&p->goal, separators);

  } else if (strcmp(first, "TotalTime") == 0) {

    p->total_time = atof(strtok(NULL, separators));

  } else if (strcmp(first, "Frequency") == 0) {

    p->frequency = atof(strtok(NULL, separators));
    p->step_size = 1./p->frequency;

  } else {

    if(alert_flag == 1) 
      fprintf(stderr, "Did not recognize '%s' parameter\n", first);
  }

  return(1);
}

/*****************************************************************************/

int read_value_one(double *value, char *separators)
{
  char *str_value;

  str_value = strtok(NULL, separators);
  if(str_value == NULL)
    return(-1);
  else
    *value = atof(str_value);

  return(1);
}

/*****************************************************************************/

int read_all_line_values(double *values, char *separators)
{
  int flag = 1;
  int count = 0;

  while(flag != -1) {
    flag = read_value_one(&values[count], separators);
    count++;
  }

  return(count - 1);
}

/*****************************************************************************/

int read_via_point(VIA_POINT *via, char *separators)
{
  int i, count;

  read_value_one(&via->time, separators);
  count = read_all_line_values(&via->joints[0], separators);
  for(i = 0; i < count; i++)
    via->joints[i] *= (M_PI/180.);
  
  return(count);
}

/*****************************************************************************/
/*****************************************************************************/

int runge_kutta_new_values(double *th_out, double *dth_out,
			   double *th, double *dth, 
			   double *k, int n, double h, double factor)
{
  int i;
  
  for(i = 0; i < 2*n; i++)
    k[i] *= h;

  for(i = 0; i < n; i++) {
    th_out[i] = th[i] + k[i]/factor;
    dth_out[i] = dth[i] + k[n+i]/factor;
  }

  return(1);
}

/*****************************************************************************/

int cubic_path(double time, VIA_POINT *start, VIA_POINT *goal,
	       double *thd, double *dthd, double *ddthd, int n)
{
  int i;
  double a0, a1, a2, a3;
  double t0, tf, dt;

  t0 = start->time;
  tf = goal->time;
  dt = tf - t0;

  if(time < t0) {
    for(i = 0; i < n; i++) {
      thd[i] = start->joints[i];
      dthd[i] = 0.;
      ddthd[i] = 0.;
    }
  } else if (time > tf) {
    for(i = 0; i < n; i++) {
      thd[i] = goal->joints[i];
      dthd[i] = 0.;
      ddthd[i] = 0.;
    }
  } else {
    for(i = 0; i < n; i++) {
      a0 = start->joints[i];
      a1 = 0.;
      a2 = 3./(dt*dt)*(goal->joints[i] - start->joints[i]);
      a3 = -2./(dt*dt*dt)*(goal->joints[i] - start->joints[i]);

      thd[i] = a0 + a1*time + a2*time*time + a3*time*time*time;
      dthd[i] = a1 + 2.*a2*time + 3.*a3*time*time;
      ddthd[i] = 2.*a2 + 6.*a3*time;
    }
  }

  return(1);
}

/*****************************************************************************/ 									
int runge_kutta(double *th, double *dth, double *tau, int n,
		int (*func)(), double h)
{
  int i, flag;
  double k1[NMAX], k2[NMAX], k3[NMAX], k4[NMAX];
  double th2[NMAX], dth2[NMAX];
  
  
  func(&k1[0], th, dth, tau);
  runge_kutta_new_values(&th2[0], &dth2[0], th, dth, &k1[0], n, h, 2.);

  func(&k2[0], &th2[0], &dth2[0], tau);
  runge_kutta_new_values(&th2[0], &dth2[0], th, dth, &k2[0], n, h, 2.);

  func(&k3[0], &th2[0], &dth2[0], tau);
  runge_kutta_new_values(&th2[0], &dth2[0], th, dth, &k3[0], n, h, 1.);
  
  func(&k4[0], &th2[0], &dth2[0], tau);
  
  for(i = 0; i < 2*n; i++) {
    k4[i] *= h;
  }
  
  for(i = 0; i < n; i++) {
    th[i]  = th[i] + k1[i]/6. + k2[i]/3. + k3[i]/3. + k4[i]/6.;
    dth[i] = dth[i] + k1[n+i]/6. + k2[n+i]/3. + k3[n+i]/3. + k4[n+i]/6.;
  }

  flag = 1;
  for(i = 0; i < n; i++)
    if(th[i] > JOINT_LIMIT || th[i] < -JOINT_LIMIT)
      flag = -1;
  
  return(flag);
}

/*****************************************************************************/

int simulate_controlled_trajectory(double *results, MANIP_PARAMETERS *p)
{
  int i, counter, flag;
  double time;
  double tau[NMAX], ddth[NMAX];
  CONTROLLER c;

  counter = 0;
  time = p->start.time;

  c.n = p->n_links;
  c.kp = &p->KP[0];
  c.kv = &p->KV[0];
  c.ki = &p->KI[0];
  c.h = p->step_size;
  c.model = p->model;

  for(i = 0; i < c.n; i++) {
    c.ierr[i] = 0.;
  }

  cubic_path(time, &p->start, &p->goal,
	     &c.th[0], &c.dth[0], &ddth[0], c.n);

  while(time < p->total_time) {
    cubic_path(time, &p->start, &p->goal,
	       &c.thd[0], &c.dthd[0], &c.ddthd[0], c.n);

    for(i = 0; i < c.n; i++) {
      c.err[i] = c.thd[i]- c.th[i];
      c.derr[i] = c.dthd[i] - c.dth[i];
    }
    
    p->control(&tau[0], &c);
    results[counter++] = time;
    
    for(i = 0; i < c.n; i++)
      results[counter++] = c.thd[i];
    for(i = 0; i < c.n; i++)
      results[counter++] = c.th[i];
    for(i = 0; i < c.n; i++)
      results[counter++] = tau[i];
    
    flag = runge_kutta(&c.th[0], &c.dth[0], &tau[0], c.n, p->robot, c.h);

    if(flag < 0)
      return(counter);
    time += c.h;
  }

  return(counter);
}

/*****************************************************************************/

int display_results(FILE *fp, double *results, int n_res, int n)
{
  int i, m;

  m = 1 + 3*n;

  for(i = 0; i < n_res; i++) {
    if((i+1)%m == 0) {
      fprintf(fp, "%.6f\n", results[i]);
    } else {
      fprintf(fp, "%.6f ", results[i]);
    }
  }

  return(1);
}

/*****************************************************************************/
